from typing import TYPE_CHECKING

import apipkg

if not TYPE_CHECKING:
    # Lazy load the package using apipkg
    apipkg.initpkg(__name__, {
        'BaseIndex': 'muzlin.index.base:BaseIndex',
        'LangchainIndex': 'muzlin.index.langchain:LangchainIndex',
        'LlamaIndex': 'muzlin.index.llama_index:LlamaIndex',
    })

else:
    # Direct imports for type checking and static analysis
    from muzlin.index.base import BaseIndex
    from muzlin.index.langchain import LangchainIndex
    from muzlin.index.llama_index import LlamaIndex



from typing import Any, List, Union

from pydantic import BaseModel


class BaseIndex(BaseModel):

    index: Union[Any, None]

    class Config:
        arbitrary_types_allowed = True

    def __call__(self, query: str) -> List[str]:
        raise NotImplementedError('Subclasses must implement this method')



import os
from enum import Enum


class EncoderDefault(Enum):
    HUGGINGFACE = {
        'embedding_model': 'BAAI/bge-small-en-v1.5',
        'language_model': 'BAAI/bge-small-en-v1.5',
    }
    FASTEMBED = {
        'embedding_model': 'BAAI/bge-small-en-v1.5',
        'language_model': 'BAAI/bge-small-en-v1.5',
    }
    OPENAI = {
        'embedding_model': os.getenv('OPENAI_MODEL_NAME', 'text-embedding-3-small'),
        'language_model': os.getenv('OPENAI_CHAT_MODEL_NAME', 'gpt-4o'),
    }
    COHERE = {
        'embedding_model': os.getenv('COHERE_MODEL_NAME', 'embed-english-v3.0'),
        'language_model': os.getenv('COHERE_CHAT_MODEL_NAME', 'command'),
    }
    VOYAGE = {
        'embedding_model': os.getenv('VOYAGE_MODEL_NAME', 'voyage-3'),
        'language_model': os.getenv('VOYAGE_CHAT_MODEL_NAME', 'voyage-3'),
    }
    MISTRAL = {
        'embedding_model': os.getenv('MISTRAL_MODEL_NAME', 'mistral-embed'),
        'language_model': os.getenv('MISTRALAI_CHAT_MODEL_NAME', 'mistral-tiny'),
    }
    AZURE = {
        'embedding_model': os.getenv('AZURE_OPENAI_MODEL', 'text-embedding-3-small'),
        'language_model': os.getenv('OPENAI_CHAT_MODEL_NAME', 'gpt-4o'),
        'deployment_name': os.getenv(
            'AZURE_OPENAI_DEPLOYMENT_NAME', 'text-embedding-3-small'
        ),
    }
    GOOGLE = {
        'embedding_model': os.getenv(
            'GOOGLE_EMBEDDING_MODEL', 'textembedding-gecko@003'
        ),
    }
    BEDROCK = {
        'embedding_model': os.environ.get(
            'BEDROCK_EMBEDDING_MODEL', 'amazon.titan-embed-image-v1'
        )
    }



import os
from time import sleep
from typing import Any, List, Optional

from pydantic import PrivateAttr

from muzlin.encoders import BaseEncoder
from muzlin.utils.defaults import EncoderDefault

# Code adapted from https://github.com/aurelio-labs/semantic-router/blob/main/semantic_router/encoders/mistral.py


class MistralEncoder(BaseEncoder):
    """Class to encode text using MistralAI."""

    _client: Any = PrivateAttr()
    _mistralai: Any = PrivateAttr()
    type: str = 'mistral'

    def __init__(
        self,
        name: Optional[str] = None,
        mistralai_api_key: Optional[str] = None,
    ):
        if name is None:
            name = EncoderDefault.MISTRAL.value['embedding_model']
        super().__init__(name=name)
        self._client, self._mistralai = self._initialize_client(
            mistralai_api_key)

    def _initialize_client(self, api_key):
        try:
            import mistralai
            from mistralai.client import MistralClient
        except ImportError:
            raise ImportError(
                'Please install MistralAI to use MistralEncoder. '
                'You can install it with: '
                '`pip install mistralai`'
            )

        api_key = api_key or os.getenv('MISTRALAI_API_KEY')
        if api_key is None:
            raise ValueError('Mistral API key not provided')
        try:
            client = MistralClient(api_key=api_key)
        except Exception as e:
            raise ValueError(
                f"Unable to connect to MistralAI {e.args}: {e}") from e
        return client, mistralai

    def __call__(self, docs: List[str]) -> List[List[float]]:
        if self._client is None:
            raise ValueError('Mistral client not initialized')
        embeds = None
        error_message = ''

        # Exponential backoff
        for _ in range(3):
            try:
                embeds = self._client.embeddings(model=self.name, input=docs)
                if embeds.data:
                    break
            except self._mistralai.exceptions.MistralException as e:
                sleep(2**_)
                error_message = str(e)
            except Exception as e:
                raise ValueError(
                    f"Unable to connect to MistralAI {e.args}: {e}") from e

        if (
            not embeds
            or not isinstance(
                embeds, self._mistralai.models.embeddings.EmbeddingResponse
            )
            or not embeds.data
        ):
            raise ValueError(
                f"No embeddings returned from MistralAI: {error_message}")
        embeddings = [embeds_obj.embedding for embeds_obj in embeds.data]
        return embeddings



import importlib.util
import os
from typing import List, Optional, Tuple, Type, Union

import joblib
import networkx as nx
import numpy as np
import torch
import torch_geometric
from pydantic import BaseModel, Field
from pygod.detector.base import Detector
from pythresh.thresholds.base import BaseThresholder
from sklearn.base import BaseEstimator, OutlierMixin, RegressorMixin
from sklearn.metrics import r2_score
from sklearn.pipeline import Pipeline
from sklearn.utils import check_array
from sklearn.utils.validation import check_is_fitted
from torch_geometric.utils import from_networkx

from muzlin.utils.logger import logger

Gtype = nx.Graph
Ttype = torch_geometric.data.Data
XType = Union[np.ndarray, List[List[Union[float, int]]]]
is_mlflow = importlib.util.find_spec('mlflow')


class GraphOutlierDetector(BaseEstimator, OutlierMixin, BaseModel):
    r"""GraphOutlierDetector class for graph based anomaly detection.

    Given a networkx graph the GraphOutlierDetector class can be
    used to fit an anomaly detection model and predict of new incoming data.

    Args:
        detector (object, optional): The PyGOD graph anomaly detection model. Defaults to PyGOD AnomalyDAE detector.
        regressor (object, optional): The regression for linking the encoded vectors with the graph anonamly likelihood scores. Defaults to RidgeCV.
        contamination (object, float, or int, optional): The level of contamation that is present in the data or unseen data. Used for setting the inlier/outlier threshold.
            - Type object: A PyThresh dynamic thresholding method can be used e.g. inter-quartile range.
            - Type float: The fraction of the fitted data that is outliers.
            - Type int: The percentile to set the threshold level e.g. 78 -> 78%, max is 100%.
            Defaults to None.
        mlflow (bool, optional): To use mlflow experiment tracking during model fitting. Setting False will fit a pickle file of the fitted model in the local folder. Defaults to True.
        model (str, optional): Name of the graph outlier detctor model to load/save. Defaults to 'outlier_detector.pkl'.
        regression_model (str, optional): Name of the vector-graph regression model to load/save. Defaults to 'regressor.pkl'.
        random_state: (int, optional): The random seed for model fitting. Defaults to 1234.

    Attributes:
        pipeline (object): The sklearn pipeline of the fitted model.
        decision_scores_ (array-like): The array of the fitted decision scores for the training data.
        threshold_ (float): The percentile used to threshold inliers from outliers in the model.
        labels_ (array-like): The array of the fitted binary labels for the training data.
        reg_R2_ (float): The R2 score of the fitted regression model on the training data.
        rm_indices_ (list): Removed node indices from the networkx graph prior to fitting. Unconnected nodes cannot be used during fitting and will be assigned outlier labels in the fitted output.

    """

    detector: Optional[Detector] = None
    regressor: Optional[RegressorMixin] = None
    contamination: Optional[Union[BaseThresholder, float, int]] = 0.1
    mlflow: Optional[bool] = True
    model: Optional[str] = 'outlier_detector.pkl'
    regressor_model: Optional[str] = 'regressor.pkl'
    random_state: Optional[int] = 1234

    pipeline: Pipeline = Field(default=None, exclude=True)
    decision_scores_: Type[np.ndarray] = Field(default=None, exclude=True)
    threshold_: float = Field(default=None, exclude=True)
    labels_: Type[np.ndarray] = Field(default=None, exclude=True)
    reg_R2_: float = Field(default=None, exclude=True)
    rm_indices_: list = Field(default=None, exclude=True)

    class Config:
        arbitrary_types_allowed = True

    def __init__(self, **data):

        super().__init__(**data)

        torch.manual_seed(self.random_state)

        self._check_is_initalized()
        if self.pipeline is not None:
            return

        if self.detector is None:

            logger.info(
                'No graph outlier detector was provided defaulting to PyGOD AnomalyDAE detector.'
            )
            global AnomalyDAE
            from pygod.detector import AnomalyDAE
            self.detector = AnomalyDAE()

        if self.regressor is None:

            logger.info(
                'No regression model was provided defaulting to sklearn RidgeCV.'
            )
            global RidgeCV
            from sklearn.linear_model import RidgeCV
            self.regressor = RidgeCV()

        self.pipeline = Pipeline([
            ('detector', self.detector)
        ])

    def fit(self, graph: Gtype, y=None):
        r"""Fit function of the OutlierDetector class.

        Args:
            graph (object): The nx.graph of the training dataset.
            y (array-like, or None, optional): Not required.
        """

        len_g = len(graph)

        # Prepare graph and vector data before fitting
        graph_torch, vectors = self._preprocess_graph(graph)

        if self.mlflow:

            if is_mlflow:
                global ml
                import mlflow as ml
                ml.autolog()
            else:
                logger.info('MLFlow not installed, defaulting to joblib')
                self.mlflow = False

        self.pipeline.fit(graph_torch)

        scores = self.pipeline.named_steps['detector'].decision_score_.numpy()
        contam = self.contamination

        # Fit mapping function
        if self.mlflow:
            ml.autolog(disable=True)
        self.regressor.fit(vectors, scores)
        pred = self.regressor.predict(vectors)
        reg_R2_ = r2_score(scores, pred)

        # Cater for different types of thresholding options
        if isinstance(self.contamination, BaseThresholder):
            lbls = self.contamination.eval(scores)
            contam = len(lbls[lbls == 0])/len(lbls)
        elif self.contamination > 1.0:
            contam /= 100
        else:
            contam = 1 - contam

        self.threshold_ = (np.percentile(scores, contam*100) if
                           contam <= 1.0 else contam * np.max(scores))

        fitted_labels = (scores > self.threshold_).astype('int').ravel()

        # Assure that the lengths of the output labels and input graph nodes match
        full_labels = np.ones(len_g)
        full_labels[np.setdiff1d(
            np.arange(len_g), self.rm_indices_)] = fitted_labels

        setattr(self.pipeline, 'threshold_', self.threshold_)
        setattr(self.pipeline, 'labels_', full_labels)
        setattr(self.pipeline, 'rm_indices_', self.rm_indices_)
        setattr(self.regressor, 'reg_R2_', reg_R2_)

        # Relog model to save attr
        if self.mlflow:
            run_id = self._fetch_mlflow_run_id()
            with ml.start_run(run_id=run_id) as _:
                ml.sklearn.log_model(self.pipeline, 'model')
                ml.sklearn.log_model(self.regressor, 'regression_model')
                ml.log_param('regressor', self.regressor)
                ml.log_metric('regressor_r2_score', reg_R2_)

        if not self.mlflow:
            joblib.dump(self.pipeline, self.model)
            joblib.dump(self.regressor, self.regressor_model)

        self._check_is_initalized()

        return self

    def predict(self, X: XType) -> np.ndarray:
        r"""Predict function of the OutlierDetector class.

        Args:
            X (array-like): The vectors to predict their labels.

        Returns:
            labels (array-like): The predicted binary labels.

        """
        check_is_fitted(self.regressor)
        X = check_array(X, ensure_2d=True)

        scores = self.decision_function(X)
        labels = (scores > self.threshold_).astype('int').ravel()
        return labels

    def decision_function(self, X: XType) -> np.ndarray:
        r"""Decision function of the OutlierDetector class.

        Args:
            X (array-like): The vectors to predict their decision scores.

        Returns:
            decision scores (array-like): The predicted decision scores.

        """
        check_is_fitted(self.regressor)
        X = check_array(X, ensure_2d=True)
        score = self.regressor.predict(X)

        return score

    def _preprocess_graph(self, graph: Gtype) -> Tuple[Ttype, XType]:

        # Check no orphaned nodes
        nodes_and_indices = [(index, node) for index, (node, degree) in
                             enumerate(dict(graph.degree()).items()) if degree == 0]

        self.rm_indices_, nodes_to_remove = zip(
            *nodes_and_indices) if nodes_and_indices else ([], [])

        graph.remove_nodes_from(nodes_to_remove)

        # Clean node attributes: retain only 'x' (first pass)
        vectors = []
        for node, attr in graph.nodes(data=True):
            if 'x' in attr:
                x = graph.nodes[node]['x']
                graph.nodes[node].clear()
                graph.nodes[node]['x'] = x
                vectors.append(x)
            else:
                raise ValueError(
                    f"Node {node} does not have the 'x' attribute.")

        graph_torch = from_networkx(graph)

        # Clean node attributes: retain only 'x' (second pass)
        attrs_to_keep = ['x', 'edge_index']
        for attr in list(graph_torch.keys()):
            if attr not in attrs_to_keep:
                delattr(graph_torch, attr)

        return graph_torch, np.array(vectors)

    def _fetch_mlflow_run_id(self):
        run = ml.last_active_run()
        run_id = run.info.run_id
        return run_id

    def _check_is_initalized(self):

        if ((os.path.exists(self.model)) & (self.pipeline is None) &
           (os.path.exists(self.regressor_model)) & (self.regressor is None)):
            self.pipeline = joblib.load(self.model)
            self.regressor = joblib.load(self.regressor_model)
        elif self.pipeline is None:
            return

        check_is_fitted(self.pipeline)
        check_is_fitted(self.regressor)

        self.threshold_ = self.pipeline.threshold_
        self.labels_ = self.pipeline.labels_
        self.rm_indices_ = self.pipeline.rm_indices_ if hasattr(
            self.pipeline, 'rm_indices_') else []
        self.decision_scores_ = self.pipeline.named_steps['detector'].decision_score_.numpy(
        )
        self.reg_R2_ = self.regressor.reg_R2_



from enum import Enum

from pydantic import BaseModel


class EncoderType(Enum):
    AZURE = 'azure'
    COHERE = 'cohere'
    OPENAI = 'openai'
    FASTEMBED = 'fastembed'
    HUGGINGFACE = 'huggingface'
    GOOGLE = 'google'
    BEDROCK = 'bedrock'
    VOYAGE = 'voyage'
    MISTRAL = 'mistral'


class EncoderInfo(BaseModel):
    name: str
    token_limit: int


class IndexType(Enum):
    LANGCHAIN = 'langchain'
    LLAMAINDEX = 'llamaindex'



import os
import time
from typing import Any, List, Optional

import requests
from pydantic import PrivateAttr

from muzlin.encoders import BaseEncoder
from muzlin.utils.logger import logger

# Code adapted from https://github.com/aurelio-labs/semantic-router/blob/main/semantic_router/encoders/huggingface.py


class HuggingFaceEncoder(BaseEncoder):
    name: str = 'BAAI/bge-small-en-v1.5'
    type: str = 'huggingface'
    tokenizer_kwargs: dict = {}
    model_kwargs: dict = {}
    device: Optional[str] = None
    _tokenizer: Any = PrivateAttr()
    _model: Any = PrivateAttr()
    _torch: Any = PrivateAttr()

    def __init__(self, **data):
        super().__init__(**data)
        self._tokenizer, self._model = self._initialize_hf_model()

    def _initialize_hf_model(self):
        try:
            import torch
        except ImportError:
            raise ImportError(
                'Please install Pytorch to use HuggingFaceEncoder.'
            )

        try:
            from transformers import AutoModel, AutoTokenizer
        except ImportError:
            raise ImportError(
                'Please install transformers to use HuggingFaceEncoder.'
            )

        self._torch = torch

        tokenizer = AutoTokenizer.from_pretrained(
            self.name,
            **self.tokenizer_kwargs,
        )

        model = AutoModel.from_pretrained(self.name, **self.model_kwargs)

        if self.device:
            model.to(self.device)

        else:
            device = 'cuda' if self._torch.cuda.is_available() else 'cpu'
            model.to(device)
            self.device = device

        return tokenizer, model

    def __call__(
        self,
        docs: List[str],
        batch_size: int = 32,
        normalize_embeddings: bool = True,
        pooling_strategy: str = 'mean',
    ) -> List[List[float]]:
        all_embeddings = []
        for i in range(0, len(docs), batch_size):
            batch_docs = docs[i: i + batch_size]

            encoded_input = self._tokenizer(
                batch_docs, padding=True, truncation=True, return_tensors='pt'
            ).to(self.device)

            with self._torch.no_grad():
                model_output = self._model(**encoded_input)

            if pooling_strategy == 'mean':
                embeddings = self._mean_pooling(
                    model_output, encoded_input['attention_mask']
                )
            elif pooling_strategy == 'max':
                embeddings = self._max_pooling(
                    model_output, encoded_input['attention_mask']
                )
            else:
                raise ValueError(
                    "Invalid pooling_strategy. Please use 'mean' or 'max'."
                )

            if normalize_embeddings:
                embeddings = self._torch.nn.functional.normalize(
                    embeddings, p=2, dim=1)

            embeddings = embeddings.tolist()
            all_embeddings.extend(embeddings)
        return all_embeddings

    def _mean_pooling(self, model_output, attention_mask):
        token_embeddings = model_output[0]
        input_mask_expanded = (
            attention_mask.unsqueeze(-1).expand(
                token_embeddings.size()).float()
        )
        return self._torch.sum(
            token_embeddings * input_mask_expanded, 1
        ) / self._torch.clamp(input_mask_expanded.sum(1), min=1e-9)

    def _max_pooling(self, model_output, attention_mask):
        token_embeddings = model_output[0]
        input_mask_expanded = (
            attention_mask.unsqueeze(-1).expand(
                token_embeddings.size()).float()
        )
        token_embeddings[input_mask_expanded == 0] = -1e9
        return self._torch.max(token_embeddings, 1)[0]


class HFEndpointEncoder(BaseEncoder):
    """
    A class to encode documents using a Hugging Face transformer model endpoint.

    Attributes:
        huggingface_url (str): The URL of the Hugging Face API endpoint.
        huggingface_api_key (str): The API key for authenticating with the Hugging Face API.
    """

    name: str = 'hugging_face_custom_endpoint'
    huggingface_url: Optional[str] = None
    huggingface_api_key: Optional[str] = None

    def __init__(
        self,
        name: Optional[str] = 'hugging_face_custom_endpoint',
        huggingface_url: Optional[str] = None,
        huggingface_api_key: Optional[str] = None,
    ):
        """
        Initializes the HFEndpointEncoder with the specified parameters.

        Args:
            name (str, optional): The name of the encoder. Defaults to
                "hugging_face_custom_endpoint".
            huggingface_url (str, optional): The URL of the Hugging Face API endpoint.
                Cannot be None.
            huggingface_api_key (str, optional): The API key for the Hugging Face API.
                Cannot be None.

        Raises:
            ValueError: If either `huggingface_url` or `huggingface_api_key` is None.
        """
        huggingface_url = huggingface_url or os.getenv('HF_API_URL')
        huggingface_api_key = huggingface_api_key or os.getenv('HF_API_KEY')

        super().__init__(name=name)  # type: ignore

        if huggingface_url is None:
            raise ValueError("HuggingFace endpoint url cannot be 'None'.")
        if huggingface_api_key is None:
            raise ValueError("HuggingFace API key cannot be 'None'.")

        self.huggingface_url = huggingface_url or os.getenv('HF_API_URL')
        self.huggingface_api_key = huggingface_api_key or os.getenv(
            'HF_API_KEY')

        try:
            self.query({'inputs': 'Hello World!', 'parameters': {}})
        except Exception as e:
            raise ValueError(
                f"HuggingFace endpoint client failed to initialize. Error: {e}"
            ) from e

    def __call__(self, docs: List[str]) -> List[List[float]]:
        """
        Encodes a list of documents into embeddings using the Hugging Face API.

        Args:
            docs (List[str]): A list of documents to encode.

        Returns:
            List[List[float]]: A list of embeddings for the given documents.

        Raises:
            ValueError: If no embeddings are returned for a document.
        """
        embeddings = []
        for d in docs:
            try:
                output = self.query({'inputs': d, 'parameters': {}})
                if not output or len(output) == 0:
                    raise ValueError('No embeddings returned from the query.')
                embeddings.append(output)

            except Exception as e:
                raise ValueError(
                    f"No embeddings returned for document. Error: {e}"
                ) from e
        return embeddings

    def query(self, payload, max_retries=3, retry_interval=5):
        """
        Sends a query to the Hugging Face API and returns the response.

        Args:
            payload (dict): The payload to send in the request.

        Returns:
            dict: The response from the Hugging Face API.

        Raises:
            ValueError: If the query fails or the response status is not 200.
        """
        headers = {
            'Accept': 'application/json',
            'Authorization': f"Bearer {self.huggingface_api_key}",
            'Content-Type': 'application/json',
        }
        for attempt in range(1, max_retries + 1):
            try:
                response = requests.post(
                    self.huggingface_url,
                    headers=headers,
                    json=payload,
                    # timeout=timeout_seconds,
                )
                if response.status_code == 503:
                    estimated_time = response.json().get('estimated_time', '')
                    if estimated_time:
                        logger.info(
                            f"Model Initializing wait for - {estimated_time:.2f}s "
                        )
                        time.sleep(estimated_time)
                        continue
                else:
                    response.raise_for_status()

            except requests.exceptions.RequestException:
                if attempt < max_retries - 1:
                    logger.info(
                        f"Retrying attempt: {attempt} for payload: {payload} ")
                    time.sleep(retry_interval)
                    retry_interval += attempt
                else:
                    raise ValueError(
                        f"Query failed with status {response.status_code}: {response.text}"
                    )

        return response.json()



import importlib.util
import os
from typing import List, Optional, Type, Union

import joblib
import numpy as np
from pydantic import BaseModel, Field
from pyod.models.base import BaseDetector
from pythresh.thresholds.base import BaseThresholder
from sklearn.base import BaseEstimator, ClassifierMixin, OutlierMixin
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler
from sklearn.utils import check_array
from sklearn.utils.validation import check_is_fitted

from muzlin.utils.logger import logger

XType = Union[np.ndarray, List[List[Union[float, int]]]]
is_mlflow = importlib.util.find_spec('mlflow')


class OutlierDetector(BaseEstimator, OutlierMixin, BaseModel):
    r"""OutlierDetector class for vector based anomaly detection.

    Given a set of embedded vectors the OutlierDetector class can be
    used to fit an anomaly detection model and predict of new incoming data.

    Args:
        detector (object, optional): The anomaly detection model. Can either be a pyod or sklearn classification model. Defaults to PyOD PCA.
        contamination (object, float, or int, optional): The level of contamation that is present in the data or unseen data. Used for setting the inlier/outlier threshold.
            - Type object: A PyThresh dynamic thresholding method can be used e.g. inter-quartile range.
            - Type float: The fraction of the fitted data that is outliers.
            - Type int: The percentile to set the threshold level e.g. 78 -> 78%, 120 -> 120%.
            Defaults to None.
        mlflow (bool, optional): To use mlflow experiment tracking during model fitting. Setting False will fit a pickle file of the fitted model in the local folder. Defaults to True.
        model (str, optional): Name of the model to load/save. Defaults to 'outlier_detector.pkl'.
        random_state: (int, optional): The random seed for model fitting. Defaults to 1234.

    Attributes:
        pipeline (object): The sklearn pipeline of the fitted model.
        decision_scores_ (array-like): The array of the fitted decision scores for the training data.
        threshold_ (float): The percentile used to threshold inliers from outliers in the model.
        labels_ (array-like): The array of the fitted binary labels for the training data.

    """

    detector: Optional[Union[BaseDetector, ClassifierMixin]] = None
    contamination: Optional[Union[BaseThresholder, float, int, None]] = None
    mlflow: Optional[bool] = True
    model: Optional[str] = 'outlier_detector.pkl'
    random_state: Optional[int] = 1234

    pipeline: Pipeline = Field(default=None, exclude=True)
    decision_scores_: Type[np.ndarray] = Field(default=None, exclude=True)
    threshold_: float = Field(default=None, exclude=True)
    Xstd_: float = Field(default=None, exclude=True)
    labels_: Type[np.ndarray] = Field(default=None, exclude=True)

    class Config:
        arbitrary_types_allowed = True

    def __init__(self, **data):

        super().__init__(**data)

        self._check_is_initalized()
        if self.pipeline is not None:
            return

        if self.detector is None:

            logger.info(
                'No outlier detector was provided defaulting to PyOD PCA detector.'
            )
            global PCA
            from pyod.models.pca import PCA
            self.detector = PCA(contamination=0.1,
                                random_state=self.random_state)

        self.pipeline = Pipeline([
            ('scaler', StandardScaler()),
            ('detector', self.detector)
        ])

    def fit(self, X: XType, y: Optional[Union[np.ndarray, list, None]] = None):
        r"""Fit function of the OutlierDetector class.

        Args:
            X (array-like): The vectors of the training dataset.
            y (array-like, or None, optional): Not required if using a PyOD model. A binary set of labels are needed if using a Sklearn classification model.

        """

        if self.mlflow:

            if is_mlflow:
                global ml
                import mlflow as ml
                ml.autolog()
            else:
                logger.info('MLFlow not installed, defaulting to joblib')
                self.mlflow = False

        X = check_array(X, ensure_2d=True)
        y = check_array(y, ensure_2d=False) if y is not None else y

        if (not np.all(np.isin(y, [0, 1]))) & (y is not None):
            raise ValueError('y should only contain binary values 0 or 1.')

        # Log training data deviation for future use
        self.pipeline.named_steps['detector'].Xstd_ = np.std(X)

        self.pipeline.fit(X, y)

        # Cater for supervised method
        if isinstance(self.pipeline.named_steps['detector'], ClassifierMixin):

            scores = self.pipeline.predict_proba(X)
            setattr(
                self.pipeline.named_steps['detector'], 'decision_scores_', scores)

        labels = self.pipeline.predict(
            X) if y is not None else self.pipeline.named_steps['detector'].labels_

        # Cater for different types of thresholding options
        if (self.contamination is not None) & (y is None):

            scores = self.pipeline.named_steps['detector'].decision_scores_
            contam = self.contamination

            if isinstance(self.contamination, BaseThresholder):
                lbls = self.contamination.eval(scores)
                contam = len(lbls[lbls == 0])/len(lbls)
            elif self.contamination > 1.0:
                contam /= 100
            else:
                contam = 1 - contam

            self.threshold_ = (np.percentile(scores, contam*100) if
                               contam <= 1.0 else contam * np.max(scores))

            labels = (scores > self.threshold_).astype('int').ravel()

        setattr(self.pipeline, 'threshold_', self.threshold_)
        setattr(self.pipeline, 'labels_', labels)

        # Relog model to save attr
        if self.mlflow:
            run_id = self._fetch_mlflow_run_id()
            with ml.start_run(run_id=run_id) as _:
                ml.sklearn.log_model(self.pipeline, 'model')

        if not self.mlflow:
            joblib.dump(self.pipeline, self.model)

        self._check_is_initalized()

        return self

    def predict(self, X: XType) -> np.ndarray:
        r"""Predict function of the OutlierDetector class.

        Args:
            X (array-like): The vectors to predict their labels.

        Returns:
            labels (array-like): The predicted binary labels.

        """
        check_is_fitted(self.pipeline)
        X = check_array(X, ensure_2d=True)

        if self.threshold_ is not None:
            scores = self.decision_function(X)
            labels = (scores > self.threshold_).astype('int').ravel()
        else:
            labels = self.pipeline.predict(X)
        return labels

    def decision_function(self, X: XType) -> np.ndarray:
        r"""Decision function of the OutlierDetector class.

        Args:
            X (array-like): The vectors to predict their decision scores.

        Returns:
            decision scores (array-like): The predicted decision scores.

        """
        check_is_fitted(self.pipeline)
        X = check_array(X, ensure_2d=True)
        return self.pipeline.decision_function(X)

    def _fetch_mlflow_run_id(self):
        run = ml.last_active_run()
        run_id = run.info.run_id
        return run_id

    def _check_is_initalized(self):

        if (os.path.exists(self.model)) & (self.pipeline is None):
            self.pipeline = joblib.load(self.model)
        elif self.pipeline is None:
            return

        check_is_fitted(self.pipeline)

        self.threshold_ = self.pipeline.threshold_
        self.labels_ = self.pipeline.labels_
        self.decision_scores_ = self.pipeline.named_steps['detector'].decision_scores_
        self.Xstd_ = self.pipeline.named_steps['detector'].Xstd_



from typing import Any, List, Optional

import numpy as np
from pydantic import PrivateAttr

from muzlin.encoders import BaseEncoder

# code adapted from https://github.com/aurelio-labs/semantic-router/blob/main/semantic_router/encoders/fastembed.py


class FastEmbedEncoder(BaseEncoder):
    type: str = 'fastembed'
    name: str = 'BAAI/bge-small-en-v1.5'
    max_length: int = 512
    cache_dir: Optional[str] = None
    threads: Optional[int] = None
    _client: Any = PrivateAttr()

    def __init__(
        self, **data
    ):
        super().__init__(**data)
        self._client = self._initialize_client()

    def _initialize_client(self):
        try:
            from fastembed import TextEmbedding
        except ImportError:
            raise ImportError(
                'Please install fastembed to use FastEmbedEncoder. '
                'You can install it with: '
                "`pip install 'muzlin[fastembed]'`"
            )

        embedding_args = {
            'model_name': self.name,
            'max_length': self.max_length,
            'cache_dir': self.cache_dir,
            'threads': self.threads,
        }

        embedding_args = {k: v for k,
                          v in embedding_args.items() if v is not None}

        embedding = TextEmbedding(**embedding_args)
        return embedding

    def __call__(self, docs: List[str]) -> List[List[float]]:
        try:
            embeds: List[np.ndarray] = list(self._client.embed(docs))
            embeddings: List[List[float]] = [e.tolist() for e in embeds]
            return embeddings
        except Exception as e:
            raise ValueError(f"FastEmbed embed failed. Error: {e}") from e



import importlib.util
import os
from collections import namedtuple
from typing import List, Optional, Type, Union

import joblib
import numpy as np
import umap
from pydantic import BaseModel, Field
from scipy.spatial.distance import cdist
from sklearn.base import BaseEstimator, ClusterMixin, OutlierMixin, TransformerMixin
from sklearn.mixture._base import BaseMixture
from sklearn.pipeline import Pipeline
from sklearn.utils import check_array
from sklearn.utils.validation import check_is_fitted

from muzlin.utils.logger import logger

XType = Union[np.ndarray, List[List[Union[float, int]]]]
is_mlflow = package_spec = importlib.util.find_spec('mlflow')

ClusterPredict = namedtuple(
    'ClusterPredict', ('nclust_cls', 'topk_cls', 'density_cls'))
ClusterDecision = namedtuple(
    'ClusterDecision', ('nclust_dev', 'topk_dev', 'density_dev'))

dim_reducer = umap.UMAP(n_components=10,
                        metric='euclidean',
                        set_op_mix_ratio=0.3,
                        min_dist=0.98,
                        local_connectivity=5,
                        random_state=1234)


# https://github.com/parthsarthi03/raptor/blob/master/raptor/cluster_utils.py

class OutlierCluster(BaseEstimator, OutlierMixin, BaseModel):
    r"""OutlierCluster class for vector based anomaly cluster detection.

    Given a set of embedded vectors the OutlierCluster class can be
    used to fit an anomaly cluster detection model and predict on new incoming data
    if the data belongs to a subcluster e.g. matched context from a RAG or not.

    Args:
        method (object, optional): The anomaly cluster detection model. Can either be a sklearn cluster or mixture model type. Defaults to sklearn KMeans.
        decomposer (object, None,  optional): The preprocessing decomposition model to reduce the dimensionality of the data to speed up clustering. Can be set to None to not be applied. Defaults to UMAP.
        n_retrieve (int, optional): Number of context new data must be compared to e.g. 10 retrieved docs per query. Defaults to 10.
        mlflow (bool, optional): To use mlflow experiment tracking during model fitting. Setting False will fit a pickle file of the fitted model in the local folder. Defaults to True.
        model (str, optional): Name of the model to load/save. Defaults to 'outlier_cluster.pkl'.
        random_state: (int, optional): The random seed for model fitting. Defaults to 1234.


    Attributes:
        pipeline (object): The sklearn pipeline of the fitted model.
        labels_ (array-like): The array of the fitted labels for clusters of the training data.
        avg_std_ (float): The average cluster centriod deviation of the training data.

    """

    method: Optional[Union[ClusterMixin, BaseMixture, None]] = None
    decomposer: Optional[Union[TransformerMixin, None]] = dim_reducer
    n_retrieve: Optional[int] = 10
    mlflow: Optional[bool] = True
    model: Optional[str] = 'outlier_cluster.pkl'
    random_state: Optional[int] = 1234

    pipeline: Pipeline = Field(default=None, exclude=True)
    avg_std_: float = Field(default=None, exclude=True)
    labels_: Type[np.ndarray] = Field(default=None, exclude=True)

    class Config:
        arbitrary_types_allowed = True

    def __init__(self, **data):

        super().__init__(**data)

        self._check_is_initalized()
        if self.pipeline is not None:
            return

        # Apply dummy function
        if self.decomposer is None:
            from sklearn.preprocessing import FunctionTransformer
            self.decomposer = FunctionTransformer(lambda x: x)

        if hasattr(self.decomposer, 'random_state'):
            self.decomposer.random_state = self.random_state

        if self.method is None:
            logger.info(
                'No clustering method was provided defaulting to KMeans.'
            )
            from sklearn.cluster import KMeans
            self.method = KMeans(n_clusters=2, random_state=self.random_state)

        self.pipeline = Pipeline([
            ('decompose', self.decomposer),
            ('cluster', self.method)
        ])

    def fit(self, X: XType, y=None):
        r"""Fit function of the OutlierCluster class.

        Args:
            X (array-like): The vectors of the training dataset.
            y (array-like, or None, optional): Not required.

        """

        if self.mlflow:

            if is_mlflow:
                global ml
                import mlflow as ml
                ml.autolog()
            else:
                logger.info('MLFlow not installed, defaulting to joblib')
                self.mlflow = False

        X = check_array(X, ensure_2d=True)

        # Set dynamic n_neighbors (applies only to umap)
        if hasattr(self.pipeline.named_steps['decompose'], 'n_neighbors'):
            n_neighbors = int(X.shape[1]/15)
            self.pipeline.named_steps['decompose'].n_neighbors = n_neighbors

        # Optimal clusters should be ~2*topk retrieved context
        optimal_n_clusters = int(X.shape[0] / (2 * self.n_retrieve))

        if hasattr(self.pipeline.named_steps['cluster'], 'n_components'):
            self.pipeline.named_steps['cluster'].n_components = optimal_n_clusters
        elif hasattr(self.pipeline.named_steps['cluster'], 'n_clusters'):
            self.pipeline.named_steps['cluster'].n_clusters = optimal_n_clusters
        else:
            pass

        labels = self.pipeline.fit_predict(X, y)

        # Get the avg median centroid deviation
        unq_labels = np.unique(labels)
        stds = [np.median(np.abs(np.mean(X[label], axis=0) - X[label]))
                for label in unq_labels]
        avg_std = np.mean(stds)

        setattr(self.pipeline, 'avg_std_', avg_std)
        setattr(self.pipeline.named_steps['cluster'], 'avg_std_', avg_std)

        # Relog model to save attr
        if self.mlflow:
            run_id = self._fetch_mlflow_run_id()
            with ml.start_run(run_id=run_id) as _:
                ml.sklearn.log_model(self.pipeline, 'model')

        if not self.mlflow:
            joblib.dump(self.pipeline, self.model)

        self._check_is_initalized()

        return self

    def predict(self, query: XType, docs: XType) -> namedtuple:
        r"""Predict function of the OutlierCluster class.

        Args:
            query (array-like): The vectors to predict their labels.
            docs (array-like): The vectors of the reference for which to compare to e.g. context

        Returns:
            labels (tuple): NamedTuple of predicted class for the three tests:
                - clust_cls: Binary label for belonging to an optimal number of clusters (e.g. not to dense or sparse in detail).
                - topk_cls: Binary label for being a realistic cluster with respect to the entire fitted data.
                - density_cls: Binary label for being from context centriod with respect to compactness.

        """

        check_is_fitted(self.pipeline)
        query = check_array(query, ensure_2d=True)
        docs = check_array(docs, ensure_2d=True)

        # Get scores and binary threshold
        scores = self.decision_function(query, docs)

        nclust_dev = scores.nclust_dev
        topk_dev = scores.topk_dev
        density_dev = scores.density_dev

        nclust_class = 0 if nclust_dev <= 0.5 else 1
        topk_class = 0 if topk_dev <= 1 else 1
        density_class = 0 if density_dev <= 1 else 1

        return ClusterPredict(nclust_cls=nclust_class, topk_cls=topk_class, density_cls=density_class)

    def decision_function(self, query: XType, docs: XType) -> namedtuple:
        r"""Decision function of the OutlierCluster class.

        Args:
            query (array-like): The vectors to predict their decision scores.
            docs (array-like): The vectors of the reference for which to compare to e.g. context

        Returns:
            scores (tuple): NamedTuple of predicted class for the three tests:
                - nclust_dev: Deviation from belonging to an optimal number of clusters (e.g. not to dense or sparse in detail).
                - topk_dev: Deviation from being a realistic cluster with respect to the entire fitted data.
                - density_dev: Deviation from context centriod with respect to compactness.

        """

        check_is_fitted(self.pipeline)
        query = check_array(query, ensure_2d=True)
        docs = check_array(docs, ensure_2d=True)

        top_k = len(docs)

        # Find the deviation from best variety response ~ 1/2 the number of requested context
        # e.g too few clusters signify dense data that may be too limited/focused
        # e.g too many clusters signify sparse data that may be too broad/general
        labels = self.get_cluster(docs)
        matched_nclust = len(np.unique(labels))
        base_k = top_k/2

        nclust_dev = abs(base_k - matched_nclust)/base_k

        # Get the mean std of all the fitted clusters
        # Assume the question is the centriod of the cluster and compare
        # the mean deviation from the psuedo-centriod to the mean std of the fitted clusters
        index_dev = self.pipeline.avg_std_
        centroid_dev = np.mean(np.abs(docs - query.squeeze()))
        topk_dev = centroid_dev/index_dev

        # Get the cosine distance from the retrived context centriod and the query
        # Compare this distance in both compactness of the cluster and zscore
        centroid = docs.mean(axis=0)

        cent_dist = np.abs(cdist([centroid], docs, metric='cosine').flatten())
        cent_q_diff = np.abs(
            cdist([centroid], [query.ravel()], metric='cosine').flatten())

        dist_range = np.max(cent_dist) - np.min(cent_dist)
        density_dev = cent_q_diff[0]/(2 * np.std(cent_dist)) * \
            (1 - np.std(cent_dist)/dist_range)

        return ClusterDecision(nclust_dev=nclust_dev, topk_dev=topk_dev, density_dev=density_dev)

    def get_cluster(self, X: XType) -> np.ndarray:
        r"""Decision function of the OutlierDetector class.

        Args:
            X (array-like): The vectors to predict their cluster label.

        Returns:
            label (array-like): The predicted cluster label for the vectors.

        """
        check_is_fitted(self.pipeline)
        X = check_array(X, ensure_2d=True)
        return self.pipeline.predict(X)

    def _fetch_mlflow_run_id(self):
        run = ml.last_active_run()
        run_id = run.info.run_id
        return run_id

    def _check_is_initalized(self):

        if (os.path.exists(self.model)) & (self.pipeline is None):
            self.pipeline = joblib.load(self.model)
        elif self.pipeline is None:
            return

        check_is_fitted(self.pipeline)

        self.avg_std_ = self.pipeline.avg_std_
        self.labels_ = self.pipeline.named_steps['cluster'].labels_



from typing import List, Optional, Union

from langchain_core.vectorstores import VectorStore, VectorStoreRetriever
from pydantic import Field

from muzlin.index import BaseIndex
from muzlin.utils.logger import logger


class LangchainIndex(BaseIndex):
    r"""LangchainIndex class for a LangChain vector based index.

    This class is used to standardize vector based LangChain Indexes for use in Muzlin.

    Args:
        index (object): A LangChain vector index e.g. FAISS. Defaults to None.
        top_k (int): Number of documents to retrieve. Defaults to 10.

    Attributes:
        retriever (object): The initialized retiever.

    """

    index: Union[VectorStore, VectorStoreRetriever] = None
    top_k: Optional[int] = 10
    retriever: VectorStoreRetriever = Field(default=None, exclude=True)

    def __init__(
        self, **data
    ):
        super().__init__(**data)

        if self.index is None:
            raise ValueError('Langchain Index is required')

        if self.top_k < 5:
            logger.warning(
                'Using less than 5 top k results may yeild suboptimal results'
            )
        if self.top_k < 1:
            raise ValueError(f"top_k needs to be >= 1, but was: {self.top_k}.")

        self.retriever:  VectorStoreRetriever = self.index.as_retriever(search_kwargs={
                                                                        'k': self.top_k})

    def __call__(self, query: str) -> List[str]:

        docs_dict = self.retriever.invoke(query)
        documents = [doc.page_content for doc in docs_dict]

        return documents



from typing import Literal, Optional, Tuple, Union

import numpy as np
from scipy.optimize import basinhopping
from scipy.stats import percentileofscore
from sklearn.metrics import matthews_corrcoef
from sklearn.utils import check_array


def negative_mcc_threshold(threshold: Union[int, float],
                           scores: Union[np.ndarray, list],
                           labels: Union[np.ndarray, list],) -> float:
    """Caculate the coresponding MCC score."""

    pred = np.zeros(len(scores), dtype=int)
    pred[scores >= threshold] = 1

    return -(matthews_corrcoef(labels, pred) + 1)


def optimize_threshold(fit_scores: Union[np.ndarray, list],
                       eval_scores: Union[np.ndarray, list],
                       labels: Union[np.ndarray, list],
                       policy: Optional[Literal['balanced',
                                                'hard', 'soft']] = 'balanced',
                       ) -> Tuple[float, float]:
    r"""Find the optimal threshold for the fitted OutlierDetector class.

    Given even a set of labeled data the optimize_threshold can be used to find the
    best threshold based on the user's criterion. This threshold can be used to
    refit the training data with the same model with the new threshold.

    Args:
        fit_scores (array-like): The array of the fitted decsion scores of the OutlierDetector class.
        eval_scores (array-like): The array of the evaluated decsion scores  from the labeled data using the OutlierDetector class.
        labels (array-like): The array of the binary labels for the threshold optimization.
        policy (str, optional): Policy type that dictates the handling of the optimization. It can be one of the following:
            - 'balanced': Uses a balanced approach by maximizing the Matthews Correlation Coefficient
            - 'hard': Uses a strict approach by setting the threshold as the minimum between the max inlier or min outlier score
            - 'soft': Uses a lenient approach by setting the threshold as the maximum between the max inlier or min outlier score
            Defaults to 'balanced'.


    Returns:
        tuple: A tuple containing two float values.
            - The first value represents thresholded decision score.
            - The second value represents the percentile of the threshold with respect to the fitted model.

    """

    fit_scores = check_array(fit_scores, ensure_2d=False)
    eval_scores = check_array(eval_scores, ensure_2d=False)
    labels = check_array(labels, ensure_2d=False)

    min_out = np.min(eval_scores[labels == 1])
    max_in = np.max(eval_scores[labels == 0])

    fit_max = np.max(fit_scores)

    if policy == 'hard':

        best_thresh = min(min_out, max_in)

    elif policy == 'soft':

        best_thresh = max(min_out, max_in)

    else:

        fit_mean = np.mean(fit_scores)
        fit_median = np.median(fit_scores)
        fit_std = np.std(fit_scores)

        # Find a global minimum option
        minimizer_kwargs = {'method': 'L-BFGS-B', 'args': (eval_scores, labels),
                            'bounds': [(fit_median, 2*fit_max)]}

        result = basinhopping(negative_mcc_threshold, niter=200,
                              x0=fit_mean + fit_std,
                              minimizer_kwargs=minimizer_kwargs,
                              stepsize=fit_std, seed=1234)

        best_thresh = result.x[0]

    if best_thresh > fit_max:
        perc = best_thresh/fit_max * 100

    else:
        perc = percentileofscore(fit_scores, best_thresh, kind='rank')

    return best_thresh, perc



import logging

import colorlog

# Code adapted from https://github.com/aurelio-labs/semantic-router/blob/main/semantic_router/utils/logger.py


class CustomFormatter(colorlog.ColoredFormatter):
    def __init__(self):
        super().__init__(
            '%(log_color)s%(asctime)s %(levelname)s %(name)s %(message)s',
            datefmt='%Y-%m-%d %H:%M:%S',
            log_colors={
                'DEBUG': 'cyan',
                'INFO': 'green',
                'WARNING': 'yellow',
                'ERROR': 'red',
                'CRITICAL': 'bold_red',
            },
            reset=True,
            style='%',
        )


def add_coloured_handler(logger):
    formatter = CustomFormatter()
    console_handler = logging.StreamHandler()
    console_handler.setFormatter(formatter)
    logger.addHandler(console_handler)
    return logger


def setup_custom_logger(name):
    logger = logging.getLogger(name)

    if not logger.hasHandlers():
        add_coloured_handler(logger)
        logger.setLevel(logging.INFO)
        logger.propagate = False

    return logger


logger: logging.Logger = setup_custom_logger(__name__)



import os
from typing import Any, List, Optional

from pydantic import PrivateAttr

from muzlin.encoders import BaseEncoder
from muzlin.utils.defaults import EncoderDefault

# Code adapted from https://github.com/aurelio-labs/semantic-router/blob/main/semantic_router/encoders/cohere.py


class VoyageAIEncoder(BaseEncoder):
    _client: Any = PrivateAttr()
    _embed_type: Any = PrivateAttr()
    type: str = 'voyageai'
    input_type: Optional[str] = 'document'

    def __init__(
        self,
        name: Optional[str] = None,
        voyage_api_key: Optional[str] = None,
        input_type: Optional[str] = 'document',
    ):
        if name is None:
            name = EncoderDefault.VOYAGE.value['embedding_model']
        super().__init__(
            name=name,
            input_type=input_type,  # type: ignore
        )
        self.input_type = input_type
        self._client = self._initialize_client(voyage_api_key)

    def _initialize_client(self, voyage_api_key: Optional[str] = None):
        """Initializes the Voyage client.

        :param voyage_api_key: The API key for the VoyageAI client, can also
        be set via the VOYAGE_API_KEY environment variable.

        :return: An instance of the VoyageAI client.
        """
        try:
            import voyageai
        except ImportError:
            raise ImportError(
                'Please install VoyageAI to use VoyageEncoder. '
                'You can install it with: '
                '`pip install voyageai`'
            )
        voyageai_api_key = voyage_api_key or os.getenv('VOYAGE_API_KEY')
        if voyageai_api_key is None:
            raise ValueError("VoyageAI API key cannot be 'None'.")
        try:
            client = voyageai.Client(voyage_api_key)
        except Exception as e:
            raise ValueError(
                f"VoyageAI API client failed to initialize. Error: {e}"
            ) from e
        return client

    def __call__(self, docs: List[str]) -> List[List[float]]:
        if self._client is None:
            raise ValueError('VoyageAI client is not initialized.')
        try:
            embeds = self._client.embed(
                texts=docs, input_type=self.input_type, model=self.name
            )
            return embeds.embeddings
        except Exception as e:
            raise ValueError(f"Voyage API call failed. Error: {e}") from e



from typing import TYPE_CHECKING

import apipkg

if not TYPE_CHECKING:
    # Lazy load the package using apipkg
    apipkg.initpkg(__name__, {
        'BaseEncoder': 'muzlin.encoders.base:BaseEncoder',
        'AzureOpenAIEncoder': 'muzlin.encoders.zure:AzureOpenAIEncoder',
        'BedrockEncoder': 'muzlin.encoders.bedrock:BedrockEncoder',
        'CohereEncoder': 'muzlin.encoders.cohere:CohereEncoder',
        'FastEmbedEncoder': 'muzlin.encoders.fastembed:FastEmbedEncoder',
        'GoogleEncoder': 'muzlin.encoders.google:GoogleEncoder',
        'HuggingFaceEncoder': 'muzlin.encoders.huggingface:HuggingFaceEncoder',
        'HFEndpointEncoder': 'muzlin.encoders.huggingface:HFEndpointEncoder',
        'MistralEncoder': 'muzlin.encoders.mistral:MistralEncoder',
        'OpenAIEncoder': 'muzlin.encoders.openai:OpenAIEncoder',
        'VoyageAIEncoder': 'muzlin.encoders.voyageai:VoyageAIEncoder',
    })

else:
    # Direct imports for type checking and static analysis
    from muzlin.encoders.base import BaseEncoder
    from muzlin.encoders.bedrock import BedrockEncoder
    from muzlin.encoders.cohere import CohereEncoder
    from muzlin.encoders.fastembed import FastEmbedEncoder
    from muzlin.encoders.google import GoogleEncoder
    from muzlin.encoders.huggingface import HFEndpointEncoder, HuggingFaceEncoder
    from muzlin.encoders.mistral import MistralEncoder
    from muzlin.encoders.openai import OpenAIEncoder
    from muzlin.encoders.voyageai import VoyageAIEncoder
    from muzlin.encoders.zure import AzureOpenAIEncoder



from typing import TYPE_CHECKING

import apipkg

if not TYPE_CHECKING:
    # Lazy load the package using apipkg
    apipkg.initpkg(__name__, {
        'OutlierCluster': 'muzlin.anomaly.cluster:OutlierCluster',
        'OutlierDetector': 'muzlin.anomaly.detector:OutlierDetector',
        'GraphOutlierDetector': 'muzlin.anomaly.graph:GraphOutlierDetector',
        'optimize_threshold': 'muzlin.anomaly.utils:optimize_threshold',
    })

else:
    # Direct imports for type checking and static analysis
    from muzlin.anomaly.cluster import OutlierCluster
    from muzlin.anomaly.detector import OutlierDetector
    from muzlin.anomaly.graph import GraphOutlierDetector
    from muzlin.anomaly.utils import optimize_threshold



from typing import List, Optional

from llama_index.core.base.base_retriever import BaseRetriever
from llama_index.core.indices.vector_store.base import VectorStoreIndex
from pydantic import Field

from muzlin.index import BaseIndex
from muzlin.utils.logger import logger


class LlamaIndex(BaseIndex):
    r"""LlamaIndex class for a Llamaindex vector based index.

    This class is used to standardize vector based Llamaindex Indexes for use in Muzlin.

    Args:
        index (object): A Llamaindex vector index e.g. VectorStoreIndex. Defaults to None.
        top_k (int): Number of documents to retrieve. Defaults to 10.

    Attributes:
        retriever (object): The initialized retiever.

    """

    index: VectorStoreIndex = None
    top_k: Optional[int] = 10
    retriever: BaseRetriever = Field(default=None, exclude=True)

    def __init__(
        self, **data
    ):
        super().__init__(**data)

        if self.index is None:
            raise ValueError('LLamaindex Index is required')

        if self.top_k < 5:
            logger.warning(
                'Using less than 5 top k results may yeild suboptimal results'
            )
        if self.top_k < 1:
            raise ValueError(f"top_k needs to be >= 1, but was: {self.top_k}.")

        self.retriever: BaseRetriever = self.index.as_retriever(
            search_kwargs={'k': self.top_k})

    def __call__(self, query: str) -> List[str]:

        nodes = self.retriever.retrieve(query)
        documents = [doc.dict()['node']['text'] for doc in nodes]

        return documents
